from django.contrib import messages

from .clone import clone_object
from .model_helpers import get_model_obj
from .models import DynamicFilterExpr
from .url_helpers import (
    referer,
    redirect_to_referer,
    redirect_to_changelist,
    redirect_to_change,
)


def dynfilters_add(request, model_name):
    try:
        model_obj = get_model_obj(model_name)
    except:
        messages.error(request, 'This type of model is unknown.')
        return redirect_to_referer(request)

    expr = DynamicFilterExpr.objects.create(
        model=model_name,
        user=request.user,
    )

    return redirect_to_change(request, expr.id, follow=True)

def dynfilters_share(request, id):
    try:
        expr = DynamicFilterExpr.objects.get(pk=id)
    except DynamicFilterExpr.DoesNotExist:
        messages.error(request, 'This filter does not exist.')
        return redirect_to_changelist(request)

    clone = clone_object(expr)
    clone.user = request.user
    clone.save()

    return redirect_to_change(request, clone.id)

def dynfilters_change(request, id):
    return redirect_to_change(request, id, follow=True)

def dynfilters_delete(request, id):
    try:
        expr = DynamicFilterExpr.objects.get(pk=id)
    except DynamicFilterExpr.DoesNotExist:
        messages.error(request, 'This filter does not exist.')
        return redirect_to_referer(request)

    expr.delete()

    return redirect_to_referer(request)



from django.apps import apps
from django.contrib import admin


def get_model_name(opts):
    return opts.model_name.capitalize()

def get_qualified_model_name(opts):
    return f'{opts.app_label}.{opts.model_name.capitalize()}'

def get_qualified_model_names(opts):
    # The model may be a proxy on a different site, so we need to consider parents also
    names = [
        get_qualified_model_name(meta)
        for parent in opts.get_parent_list() 
        if (meta := parent._meta)
    ]

    names.append(
        get_qualified_model_name(opts)
    )

    return names

def get_model_obj(qmodel_name):
    app_label, model_name = qmodel_name.split('.')
    return apps.get_model(app_label, model_name)

def get_model_admin(obj):
    model_obj = get_model_obj(obj.model)
    return admin.site._registry.get(model_obj)

def get_model_choices():
    return [
        (
            get_qualified_model_name(opts),
            get_model_name(opts)
        )
        for model_obj in apps.get_models()
        if has_dynfilter(model_obj, (opts := model_obj._meta))
    ]


def has_dynfilter(model_obj, opts):
    model_admin = admin.site._registry.get(model_obj)
    return hasattr(model_admin, 'dynfilters_fields') and not opts.proxy

def get_dynfilters_fields(model_admin):
    def humanize(f):
        if f == '-':
            return ('-', '---------')

        if isinstance(f, str):
            return (
                f, 
                ' OR '.join([
                    s.replace('__', ' ').replace('_', ' ').capitalize() 
                    for s in f.split('|')
                ])
            )

        return f

    fields = getattr(model_admin, 'dynfilters_fields', [])
    return [humanize(f) for f in fields]

def get_dynfilters_select_related(model_admin):
    return getattr(model_admin, 'dynfilters_select_related', [])

def get_dynfilters_prefetch_related(model_admin):
    return getattr(model_admin, 'dynfilters_prefetch_related', [])



# https://stackoverflow.com/a/61729857/2893408
# https://github.com/morlandi/test-django-clone

def clone_object(obj, attrs={}):
    # we start by building a "flat" clone
    clone = obj._meta.model.objects.get(id=obj.id)
    clone.id = None

    # if caller specified some attributes to be overridden, use them
    for key, value in attrs.items():
        setattr(clone, key, value)

    # save the partial clone to have a valid ID assigned
    clone.save()

    # Scan field to further investigate relations
    fields = clone._meta.get_fields()
    for field in fields:

        # Manage M2M fields by replicating all related records found on parent "obj" into "clone"
        if not field.auto_created and field.many_to_many:
            for row in getattr(obj, field.name).all():
                getattr(clone, field.name).add(row)

        # Manage 1-N and 1-1 relations by cloning child objects
        if field.auto_created and field.is_relation:
            if field.many_to_many:
                # do nothing
                pass
            else:
                # provide "clone" object to replace "obj" on remote field
                attrs = {
                    field.remote_field.name: clone
                }
                children = field.related_model.objects.filter(**{field.remote_field.name: obj})
                for child in children:
                    clone_object(child, attrs)

    return clone



from django.urls import path, reverse

from . import views

urlpatterns = [
    path('<str:model_name>/add/', views.dynfilters_add, name='dynfilters_add'),
    path('<int:id>/share/', views.dynfilters_share, name='dynfilters_share'),
    path('<int:id>/change/', views.dynfilters_change, name='dynfilters_change'),
    path('<int:id>/delete/', views.dynfilters_delete, name='dynfilters_delete'),
]



from operator import itemgetter

from django import forms
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError

from adminsortable2.admin import CustomInlineFormSet

from .models import DynamicFilterExpr, DynamicFilterTerm
from .utils import str_as_date, str_as_date_range


class DynamicFilterExprForm(forms.ModelForm):
    class Meta:
        model = DynamicFilterExpr
        fields = ('name', 'is_global')


class DynamicFilterTermInlineFormSet(CustomInlineFormSet):
    def clean(self):
        i = 0

        for form in self.forms:
            # Skip if the inline has been marked as deleted
            _del = form.cleaned_data.get('DELETE')
            if _del:
                continue

            op = form.cleaned_data.get('op')
            if op == '(':
                i += 1
            elif op == ')':
                i -= 1

            if i < 0:
                raise ValidationError("Missing opening parenthesis")

        if i:
            raise ValidationError("Missing closing parenthesis")


class DynamicFilterTermInlineForm(forms.ModelForm):
    class Meta:
        model = DynamicFilterTerm
        fields = ('op', 'field', 'lookup', 'value', 'order')

    def clean(self):
        errors = {}

        op, field, lookup, value = itemgetter('op', 'field', 'lookup', 'value')(self.cleaned_data)

        if op in ('-', '!'):
            if field == '-':
                errors.update({'field': 'Missing value'})

            if lookup == '-':
                errors.update({'lookup': 'Missing value'})

            if not value:
                if lookup not in ('isnull', 'isnotnull', 'istrue', 'isfalse'):
                    errors.update({'value': 'Missing value'})

            else:
                if 'date' in field:
                    if lookup == 'range':
                        try:
                            str_as_date_range(value)
                        except:
                            errors.update({'value': 'Should be "DD/MM/YYYY, DD/MM/YYYY"'})

                    else:
                        try:
                            str_as_date(value)
                        except:
                            errors.update({'value': 'Should be "DD/MM/YYYY"'})


                if lookup in ('year', 'month', 'day', 'lt', 'gt', 'lte', 'gte'):
                    try:
                        float(value)
                    except:
                        errors.update({'value': 'Should be a number'})

        else:
            pass # will be handled by model clean()

        if errors:
            raise ValidationError(errors)



from furl import furl

from django.shortcuts import redirect
from django.urls import reverse
from django.utils.http import urlencode


def referer(request):
    return request.META["HTTP_REFERER"]

def redirect_to_referer(request):
    return redirect(referer(request))

def redirect_to_changelist(request):
    return redirect(reverse('admin:dynfilters_dynamicfilterexpr_changelist'))

def redirect_to_change(request, id, follow=False):
    url = reverse('admin:dynfilters_dynamicfilterexpr_change', args=(id,))

    if follow:
        query_string = urlencode({
            'next': (
                furl(referer(request))
                    .remove(['filter'])
                    .add({'filter': id})
                    .url
            ),
        })

        return redirect(f'{url}?{query_string}')

    return redirect(url)

def redirect_to_referer_next(request, response):
    _next = (
        furl(referer(request))
            .args
            .get('next')
    )

    if _next:
        return redirect(_next)

    return response



from django.db.models import Q

# Function that returns value of
# expression after evaluation.
# Adapted from https://www.geeksforgeeks.org/expression-evaluation/
def evaluate(tokens):
    # Function to find precedence 
    # of operators.
    def precedence(op): 
        if op == ' ': return 3 # no-op
        if op == '&': return 2 # AND
        if op == '|': return 1 # OR
        return 0
     
    # Function to perform arithmetic 
    # operations.
    def popAndApplyOp():
        op = ops.pop()

        # no-op
        if op == ' ': return Q()

        b = values.pop()
        a = values.pop()

        # AND | OR
        if op == '&': return a & b
        if op == '|':  return a | b

    # stack to store Q objects.
    values = []

    # stack to store operators.
    ops = []

    for token in tokens:

        # Current token is an opening 
        # brace, push it to 'ops'
        if token.op == '(':
            ops.append(token.op)

        # Current token is a term, push 
        # it to stack for Q objects.
        elif token.op == "-" or token.op == '!':
            values.append(token.as_q())
          
        # Closing brace encountered, 
        # solve entire brace.
        elif token.op == ')':
            while ops and ops[-1] != '(':
                values.append(popAndApplyOp())
            
            # pop opening brace.
            ops.pop()
        
        # Current token is an operator.
        else:
            # While top of 'ops' has same or 
            # greater precedence to current 
            # token, which is an operator.
            # Apply operator on top of 'ops' 
            # to top two elements in values stack.
            while (ops and
                precedence(ops[-1]) >=
                precedence(token.op)):
                        
                values.append(popAndApplyOp())
            
            # Push current token to 'ops'.
            ops.append(token.op)
        
    # Entire expression has been parsed 
    # at this point, apply remaining ops 
    # to remaining values.
    while ops:            
        values.append(popAndApplyOp())

    # Top of 'values' contains result, 
    # return it.
    return values[-1]



# Generated by Django 4.0.5 on 2022-06-03 13:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dynfilters', '0002_alter_dynamicfilterexpr_model'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dynamicfilterexpr',
            name='name',
            field=models.CharField(default='Report', max_length=128),
        ),
    ]



from functools import reduce
from operator import or_

from django.apps import apps
from django.contrib.auth.models import User
from django.db import connection
from django.db import models
from django.db.models import Q
from django.db.models.deletion import CASCADE, SET_NULL
from django.db.models.sql.compiler import SQLCompiler
from django.db.models.sql.query import Query
from django.utils.translation import gettext_lazy as _

from . import shunting_yard
from .model_helpers import get_model_obj
from .utils import (
    previous,
    str_as_date, 
    str_as_date_range, 
)


class DynamicFilterExpr(models.Model):
    class Meta:
        verbose_name = 'Filter'
        verbose_name_plural = 'Filters'

    name = models.CharField(max_length=128, default='Report')
    model = models.CharField(max_length=64, db_index=True)
    user = models.ForeignKey(User, on_delete=models.CASCADE, db_index=True)
    is_global = models.BooleanField(default=False, db_index=True, help_text='Make filter accessible to all.')

    def __str__(self):
        return self.name

    # Make implicit operators explicit, to ensure the ops stack is never empty.
    def normalized_terms(self):
        nterms = []

        for prev, item in previous(self.dynamicfilterterm_set.all()):
            if prev:
                # Add implicit ANDs.
                if prev.op in ('-', '!', ')') and item.op in ('-', '!', '('):
                    nterms.append(DynamicFilterTerm(op='&'))

                # Add no-ops.
                elif prev.op in ('(', '&', '|') and item.op in (')', '&', '|'):
                    nterms.append(DynamicFilterTerm(op=' '))

            nterms.append(item)

        if not nterms:
            # Add no-op to avoid an empty filter.
            nterms.append(DynamicFilterTerm(op=' '))

        return nterms

    def as_q(self):
        terms = self.normalized_terms()
        return shunting_yard.evaluate(terms)

    def as_sql(self):
        model_obj = get_model_obj(self.model)
        query = Query(model_obj)

        where = self.as_q().resolve_expression(query)
        
        compiler = query.get_compiler(using='default')
        query_str, args = where.as_sql(compiler, connection)

        return query_str % tuple(args)


class DynamicFilterTerm(models.Model):
    OP_CHOICES = [
        ('-', '-'),
        ('!', 'NOT'),
        ('&', 'AND'),
        ('|', 'OR'),
        ('(', '('),
        (')', ')'),
    ]

    LOOKUP_CHOICES = [
        ('-', '---------'),
        ('=', 'Equals'),
        ('icontains', 'Contains'),
        ('istartswith', 'Starts with'),
        ('iendswith', 'Ends with'),
        ('in', 'One of'),
        # ('iregex', 'One of'),
        ('-', '---------'),
        ('range', 'Date Range'),
        ('year', 'Date Year'),
        ('month', 'Date Month'),
        ('day', 'Date Day'),
        ('-', '---------'),
        ('isnull', 'Is NULL'),
        ('isnotnull', 'Is not NULL'),
        ('istrue', 'Is TRUE'),
        ('isfalse', 'Is FALSE'),
        ('-', '---------'),
        ('lt', 'Less Than'),
        ('gt', 'Greater Than'),
        ('lte', 'Less Than or Equal To'),
        ('gte', 'Greater Than or Equal To'),
    ]

    class Meta:
        ordering = ('order',)
        verbose_name = 'Field'
        verbose_name_plural = 'Fields'

    filter = models.ForeignKey(DynamicFilterExpr, on_delete=CASCADE)
    op = models.CharField(max_length=1, choices=OP_CHOICES, default='-')
    field = models.CharField(max_length=64, blank=True, null=True)
    lookup = models.CharField(max_length=16, choices=LOOKUP_CHOICES, default='-')
    value = models.CharField(max_length=100, blank=True, null=True)
    bilateral = models.BooleanField(default=False)
    order = models.PositiveSmallIntegerField(default=0, blank=True, db_index=True)

    @property
    def fields(self):
        return self.field.split('|')

    def __str__(self):
        if self.op in ('-', '!'):
            expr = ' OR '.join([
                f'{self.get_keypath(field)} {self.get_operator()} {self.get_value()}'
                for field in self.fields
            ])

            return (
                f'NOT({expr})'
                if self.op == '!' else
                expr
            )

        return self.op

    def clean(self):
        if self.op in ('&', '|', '(', ')'):
            # Clear irrelevant fields for these operators
            self.field = '-'
            self.lookup = '-'
            self.value = None

        elif self.op in ('-', '!'):
            # Clear irrelevant fields for these operators
            if self.lookup in ('isnull', 'isnotnull', 'istrue', 'isfalse'):
                self.value = None

    def get_operator(self):
        if self.op == '!':
            return '!='
        return '=='

    def get_keypath(self, field):
        if self.lookup in ('=', 'istrue', 'isfalse'):
            return field

        if self.lookup in ('isnull', 'isnotnull'):
            return f'{field}__isnull'

        return f'{field}__{self.lookup}'

    def get_value(self):
        if self.lookup in ('isnull', 'istrue'):
            return True

        if self.lookup in ('isnotnull', 'isfalse'):
            return False

        if self.lookup == 'in':
            return self.value.split(',')

        if self.lookup == 'range':
            return str_as_date_range(self.value)

        if 'date' in self.field:
            return str_as_date(self.value)

        return self.value

    def get_term(self, field):
        return {self.get_keypath(field): self.get_value()}

    def as_q(self):
        q = reduce(or_, [
            Q(**self.get_term(field))
            for field in self.fields
        ])

        return (
            ~q
            if self.op == '!' else
            q
        )


class DynamicFilterColumn(models.Model):
    class Meta:
        ordering = ('order',)
        verbose_name = 'Column'
        verbose_name_plural = 'Columns'

    filter = models.ForeignKey(DynamicFilterExpr, on_delete=CASCADE)
    field = models.CharField(max_length=64, blank=True, null=True)
    order = models.PositiveSmallIntegerField()

    def __str__(self):
        return self.field


class DynamicFilterColumnSortOrder(models.Model):
    class Meta:
        ordering = ('order',)
        verbose_name = 'Column sort order'
        verbose_name_plural = 'Column sort orders'

    filter = models.ForeignKey(DynamicFilterExpr, on_delete=CASCADE)
    field = models.CharField(max_length=64, blank=True, null=True)
    order = models.PositiveSmallIntegerField()

    def __str__(self):
        return self.field



from itertools import tee, chain
import collections
import datetime
import itertools


def str_as_date(value):
    return datetime.datetime.strptime(value, '%d/%m/%Y').date()

def str_as_date_range(value):
    d1, d2 = value.split(',')
    d1, d2 = d1.strip(), d2.strip()

    return [
        datetime.datetime.strptime(d1, '%d/%m/%Y').date(),
        datetime.datetime.strptime(d2, '%d/%m/%Y').date(),
    ]


def previous(some_iterable):
    prevs, items = tee(some_iterable, 2)
    prevs = chain([None], prevs)
    return zip(prevs, items)


def flatten(iterable, ltypes=collections.Iterable):
    remainder = iter(iterable)
    while True:
        try:
            first = next(remainder)
        except StopIteration:
            return
        if isinstance(first, ltypes) and not isinstance(first, (str,)):
            remainder = itertools.chain(first, remainder)
        else:
            yield first


def to_int(value):
    if isinstance(value, (int, float)):
        return int(value)

    if isinstance(value, str) and value.isdigit():
        return int(value)

    if isinstance(value, (list, tuple)):
        for s in value:
            if s.isdigit():
                return int(s)

    return None
